from confluent_kafka import SerializingProducer
from confluent_kafka.schema_registry.avro import AvroSerializer
from fluvii.general_utils import parse_headers, Admin
from fluvii.exceptions import ProducerTimeoutFailure
import json
import mmh3
import uuid
import logging

LOGGER = logging.getLogger(__name__)


class Producer:
    def __init__(self, urls, schema_registry=None, topic_schema_dict=None, metrics_manager=None, client_auth_config=None, settings_config=None):
        self._urls = ','.join(urls) if isinstance(urls, list) else urls
        self._auth = client_auth_config
        self._settings = settings_config
        self._producer = None
        self._topic_partition_metadata = {}
        self._schema_registry = schema_registry
        self._admin = None

        self.topic_schemas = {}
        self.metrics_manager = metrics_manager

        self._init_producer()

        if not topic_schema_dict:
            topic_schema_dict = {}
        for topic, schema in topic_schema_dict.items():
            self.add_topic(topic, schema)

    def __getattr__(self, attr):
        """Note: this includes methods as well!"""
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            return self._producer.__getattribute__(attr)

    @staticmethod
    def _callback(error, message):
        """
        Logs the returned message from the Broker after producing
        NOTE: Headers not supported on the message for callback for some reason.
        NOTE: Callback requires these args
        """
        if error:
            LOGGER.critical(error)
        else:
            LOGGER.debug(f'Message produced successfully!')

    def _make_config(self):
        settings = {
            "bootstrap.servers": self._urls,

            "on_delivery": self._callback,
            "enable.idempotence": "true",
            "acks": "all",

            # Registry Serialization Settings
            "key.serializer": AvroSerializer(self._schema_registry, schema_str='{"type": "string"}'),
            "value.serializer": lambda: None,
        }

        if self._settings:
            settings.update(self._settings.as_client_dict())
        if self._auth:
            settings.update(self._auth.as_client_dict())
        return settings

    def _init_producer(self):
        LOGGER.info('Initializing producer...')
        self._producer = SerializingProducer(self._make_config())
        self._producer.poll(0)  # allows topic metadata querying to work immediately after init
        LOGGER.info('Producer initialized successfully!')

    def _get_topic_metadata(self, topic):
        partitions = self._producer.list_topics().topics[topic].partitions
        LOGGER.debug(partitions)
        self._topic_partition_metadata.update({topic: len(partitions)})

    def _get_topic_partition_count(self, topic):
        try:
            return self._topic_partition_metadata[topic]
        except KeyError:
            self._get_topic_metadata(topic)
            return self._topic_partition_metadata[topic]

    def _partitioner(self, key, topic):
        return mmh3.hash(key) % self._get_topic_partition_count(topic)

    def _generate_guid(self):
        return str(uuid.uuid1())

    def _add_serializer(self, topic, schema):
        LOGGER.info(f'Adding serializer for producer topic {topic}')
        self.topic_schemas.update({topic: AvroSerializer(self._schema_registry, json.dumps(schema))})

    def add_topic(self, topic, schema, overwrite=False):
        """For adding topics at runtime"""
        if topic in self.topic_schemas and not overwrite:
            LOGGER.debug("producer already has a schema registered for that topic...")
        else:
            LOGGER.info(f'Adding schema for topic {topic}')
            self._get_topic_metadata(topic)
            self._add_serializer(topic, schema)

    def _format_produce(self, value, key, topic, headers, partition, message_passthrough):
        headers_out = {}
        if message_passthrough:
            headers_out = parse_headers(message_passthrough.headers())
            if not key:
                key = message_passthrough.key()

        headers_out.update(headers if headers else {})
        headers_out = {key: value for key, value in headers_out.items() if value is not None}
        if 'guid' not in headers_out:
            headers_out['guid'] = self._generate_guid()

        if not topic:
            topics = [topic for topic in self.topic_schemas if '__changelog' not in topic]
            if len(topics) == 1:
                topic = topics[0]
            else:
                raise Exception('Topic must be defined if managing more than 1 topic')
        if partition is None:
            partition = self._partitioner(key, topic)
        self._producer._value_serializer = self.topic_schemas[topic]
        if '__changelog' not in topic:  # TODO: add a separate logger for changelog stuff, but for now it just clutters things
            LOGGER.debug(f'Adding message to the produce queue for [topic, partition, key] - [{topic}, {partition}, {repr(key)}]')
            LOGGER.info(f'Producing message with guid {headers_out["guid"]}')
        return dict(topic=topic, key=key, value=value, headers=headers_out, partition=partition)

    def produce(self, value, key=None, topic=None, headers=None, partition=None, message_passthrough=None):
        produce_dict = self._format_produce(value, key, topic, headers, partition, message_passthrough)
        self._producer.poll(0)
        self._producer.produce(**produce_dict)
        if self.metrics_manager:
            self.metrics_manager.inc_metric('messages_produced', label_dict={'topic': produce_dict['topic']})

    def _confirm_produce(self, attempts=3, timeout=20):
        """
        Ensure that messages are actually produced by forcing synchronous processing. Must manually check the events queue
        and see if it's truly empty since flushing timeouts do not actually raise an exception for some reason.

        NOTE: Only used for synchronous producing, which is dramatically slower than asychnronous.
        """
        attempt = 1
        LOGGER.debug("Sending/confirming the leftover messages in producer message queue")
        while self._producer.__len__() > 0:
            if attempt <= attempts:
                LOGGER.debug(f"Produce flush attempt: {attempt} of {attempts}")
                self._producer.flush(timeout=timeout)
                attempt += 1
            else:
                raise ProducerTimeoutFailure

    def close(self):
        self._confirm_produce()


class TransactionalProducer(Producer):
    def __init__(self, urls, transactional_id, **kwargs):
        self._transactional_id = transactional_id
        self.active_transaction = False
        super().__init__(urls, **kwargs)

    def _make_config(self):
        config = super()._make_config()
        config.update({"transactional.id": self._transactional_id})
        return config

    def _init_producer(self):
        super()._init_producer()
        self._producer.init_transactions()

    def begin_transaction(self, *args, **kwargs):
        LOGGER.debug('Initializing a transaction...')
        self._producer.begin_transaction(*args, **kwargs)
        self.active_transaction = True
    
    def produce(self, value, key=None, topic=None, headers=None, partition=None, message_passthrough=None):
        if not self.active_transaction:
            self.begin_transaction()
        super().produce(value, key=key, topic=topic, headers=headers, partition=partition, message_passthrough=message_passthrough)
        
    def abort_transaction(self, *args, **kwargs):
        LOGGER.debug('Aborting the transaction...')
        self._producer.abort_transaction(*args, **kwargs)
        self.active_transaction = False
        
    def commit_transaction(self, *args, **kwargs):
        self._producer.commit_transaction(*args, **kwargs)
        self._producer.poll(0)
        self.active_transaction = False

    def close(self):
        pass



from confluent_kafka import TopicPartition, KafkaException
from fluvii.sqlite import SqliteFluvii
from fluvii.exceptions import TransactionNotRequired
from datetime import datetime
import logging
from time import sleep

LOGGER = logging.getLogger(__name__)


# TODO: override the "Partition" class types to match TopicPartition so we dont need to do ._partition calls all the time

class Partition:
    def __init__(self, topic=None, partition=None, offset=0, p_obj=None):
        if p_obj:
            self._partition = p_obj
        else:
            self._partition = TopicPartition(topic=topic, partition=partition, offset=offset)
        self.assigned = False
        self.active = False

    def __getattr__(self, attr):
        """Note: this includes methods as well!"""
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            return self._partition.__getattribute__(attr)

    def mark_assigned(self):
        self.assigned = True
        self.active = True


class TablePartition(Partition):
    def __init__(self, topic=None, partition=None, offset=0, p_obj=None):
        super().__init__(topic=topic, partition=partition, offset=offset, p_obj=p_obj)
        self.table_assigned = False
        self.is_recovering = False
        self.table_offset = None
        self.lowwater = None
        self.highwater = None

    def __getattr__(self, attr):
        """Note: this includes methods as well!"""
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            return self._partition.__getattribute__(attr)

    @property
    def needs_recovery(self):
        # note: >2 is because: +1 due to table tracking "current" offset, and +1 due to last offset being a transaction marker
        has_consumable_offets = self.lowwater != self.highwater
        table_is_behind = self.recovery_offset_delta > 2
        LOGGER.debug(f'table {self.partition}: has consumable offsets? {has_consumable_offets}; table is behind? {table_is_behind}')
        return has_consumable_offets and table_is_behind

    @property
    def recovery_offset_delta(self):
        return self.highwater - self.table_offset

    def refresh_recovery_status(self):
        if self.is_recovering and not self.needs_recovery:
            self.is_recovering = False


class TableRebalanceManager:
    """
    Helps manage rebalancing with tables. In summary, will:
    1. assign both the "primary" topic and the changelog topic (which is tied to a table) upon a rebalance.
    2. immediately pause ALL partition consumption.
    3. check table offsets against related changelog partition highwaters; seek and resume any with deltas.
    4. If any changelogs resumed, consume from all unpaused partitions until tables are fully recovered.
    5. Pause all changelog partitions indefinitely, resume all primary partitions.

    """
    def __init__(self, consumer, changelog_topic, tables, config):
        self.consumer = consumer
        self.changelog = changelog_topic
        self._primary_partitions = []
        self._changelog_partitions = []
        self.tables = tables
        self._config = config
        self._init_recovery_time_remaining_attrs()
        LOGGER.debug('table rebalance manager initialized')

    @property
    def _all_partitions(self):
        return self._primary_partitions + self._changelog_partitions

    @property
    def _active_partitions(self):
        return [p for p in self._all_partitions if p.active]

    @property
    def recovery_partitions(self):
        return [p for p in self._changelog_partitions if p.is_recovering]

    def _resume_partitions(self, partitions):
        self.consumer.resume([p._partition for p in partitions])
        for p in partitions:
            p.active = True
            
    def _pause_all_changelog_partitions(self):
        LOGGER.debug('Pausing all active partitions')
        self.consumer.pause([p._partition for p in self._changelog_partitions if p.active])
        for p in self._changelog_partitions:
            p.active = False

    def _pause_all_active_partitions(self):
        LOGGER.debug('Pausing all active partitions')
        self.consumer.pause([p._partition for p in self._active_partitions])
        for p in self._active_partitions:
            p.active = False
    
    def _init_table_dbs(self):
        for p in self._changelog_partitions:
            if not p.table_assigned:
                self.tables[p.partition] = SqliteFluvii(f'p{p.partition}', self._config)
                p.table_assigned = True

    def _init_recovery_time_remaining_attrs(self):
        self._recovery_time_start = int(datetime.timestamp(datetime.now()))
        self._recovery_offsets_remaining = 0
        self._recovery_offsets_handled = 0

    def _get_changelog_watermarks(self):
        for p in self._changelog_partitions:
            p.lowwater, p.highwater = self.consumer.get_watermark_offsets(p._partition, timeout=10)

    def _get_table_offsets(self):
        for p in self._changelog_partitions:
            p.table_offset = self.tables[p.partition].offset
            if p.table_offset < p.lowwater:
                LOGGER.debug('Adjusting table offset due to it being lower than the changelog lowwater')
                p.table_offset = p.lowwater - 1  # -1 since the table marks the latest offset it has, not which offset is next
        LOGGER.info(f'table offsets and highwaters (t,o,h) are : {[(p.partition, p.table_offset, p.highwater) for p in self._changelog_partitions]}')

    def _set_partition_recovery_statuses(self):
        for p in self._changelog_partitions:
            if p.needs_recovery:
                p.is_recovering = True

    def _changelog_seek_and_activate(self):
        LOGGER.debug('Seeking all changelog partitions that need to be recovered from...')
        for p in self.recovery_partitions:
            try:
                self.consumer.seek(TopicPartition(topic=p.topic, partition=p.partition, offset=p.table_offset))
            except KafkaException as kafka_error:
                if 'Failed to seek to offset' in kafka_error.args[0].str():
                    LOGGER.debug('Running a consumer poll to allow seeking to work on the changelog partitions...')
                    try:
                        self.consumer.consume(timeout=8)
                    except TransactionNotRequired:
                        pass
                    self._changelog_seek_and_activate()
                else:
                    raise
        self._resume_partitions(self.recovery_partitions)

    def _update_recovery_partition_table_offsets(self):
        for p in self.recovery_partitions:
            p.table_offset = self.tables[p.partition].offset
            
    def _set_all_tables_to_latest_offset(self):
        for p in self._changelog_partitions:
            if p.table_offset < p.highwater:
                LOGGER.debug(f'Setting table p{p.partition} to {p.highwater} to mark as fully recovered')
                self.tables[p.partition].set_offset(p.highwater)
                self.tables[p.partition].commit()
    
    def _close_tables(self, partitions=None):
        interrupt = None
        if not partitions:
            partitions = list(self.tables.keys())
        LOGGER.debug(f'Table - closing connections for partitions {partitions}')
        for p in partitions:
            try:
                self.tables[p].close()
                sleep(.1)
                LOGGER.debug(f'p{p} table connection closed.')
                del self.tables[p]
            except KeyError:
                LOGGER.debug(
                    f'Table p{p} did not seem to be mounted and thus could not unmount,'
                    f' likely caused by multiple rebalances in quick succession.'
                    f' This is unliklely to cause issues as the client is in the middle of adjusting itself, '
                    f' but should be noted.')
            except Exception as e:
                LOGGER.debug(f'Interrupt received during table db closing, {e}')
                interrupt = e
        LOGGER.info(f'Table - closed connections for partitions {partitions}')
        if interrupt:  # ensure all table cleanup happens
            LOGGER.info('Continuing with exception interrupt raised during table closing')
            raise interrupt
    
    def assign_partitions(self, partition_obj_list):
        LOGGER.info(f'ASSIGNMENT called for: {partition_obj_list}')
        new_primary = [Partition(p_obj=p) for p in partition_obj_list]
        new_changelog = [TablePartition(topic=self.changelog, partition=p.partition) for p in partition_obj_list]
        all_new_partitions = new_primary + new_changelog
        LOGGER.debug(f'Assigning partitions to consumer...')
        self.consumer.incremental_assign([p._partition for p in all_new_partitions])
        LOGGER.debug(f'partitions assigned!')
        self._primary_partitions += new_primary
        self._changelog_partitions += new_changelog
        LOGGER.debug('marking all partitions as assigned...')
        for p in all_new_partitions:
            p.mark_assigned()
        self._pause_all_active_partitions()
        self._init_table_dbs()
        LOGGER.debug(f'assignment steps finished!')

    def unassign_partitions(self, partition_obj_list):
        LOGGER.info(f'UNASSIGNMENT called for: {partition_obj_list}')
        partitions = [p.partition for p in partition_obj_list]
        self.consumer.incremental_unassign([p._partition for p in self._all_partitions if p.partition in partitions])
        self._primary_partitions = [obj for obj in self._primary_partitions if obj.partition not in partitions]
        self._changelog_partitions = [obj for obj in self._changelog_partitions if obj.partition not in partitions]
        self._close_tables(partitions=partitions)

    def set_table_and_recovery_state(self):
        self._init_recovery_time_remaining_attrs()
        self._get_changelog_watermarks()
        self._get_table_offsets()
        self._set_partition_recovery_statuses()
        if self.recovery_partitions:
            self._changelog_seek_and_activate()
            self._recovery_offsets_remaining = sum([p.recovery_offset_delta for p in self.recovery_partitions])

    def update_recovery_status(self):
        self._update_recovery_partition_table_offsets()
        for partition in self.recovery_partitions:
            partition.refresh_recovery_status()
        recovery_elapsed_time_secs = max(int(datetime.timestamp(datetime.now())) - self._recovery_time_start, 1)
        offsets_remaining = sum([p.recovery_offset_delta for p in self.recovery_partitions])
        offsets_processed = self._recovery_offsets_remaining - offsets_remaining
        self._recovery_offsets_handled += offsets_processed
        self._recovery_offsets_remaining = offsets_remaining
        recovery_rate_offsets_secs = (max(self._recovery_offsets_handled, 1)) / recovery_elapsed_time_secs  # avoid 0 div error
        remaining_time_mins = round((self._recovery_offsets_remaining / recovery_rate_offsets_secs) / 60, 2)
        LOGGER.debug(f'RECOVERY - offsets processed this iteration: {offsets_processed}')
        LOGGER.debug(f'RECOVERY - current process rate: {int(recovery_rate_offsets_secs)} offsets/sec')
        LOGGER.info(f'RECOVERY - {self._recovery_offsets_handled} TOTAL OFFSETS PROCESSED OVER {round(recovery_elapsed_time_secs / 60, 2)} minutes')
        LOGGER.info(f'RECOVERY - ESTIMATED TIME REMAINING: {remaining_time_mins} minutes')
        LOGGER.info(f'RECOVERY - OFFSETS REMAINING: {self._recovery_offsets_remaining}')

    def finalize_rebalance(self):
        self._set_all_tables_to_latest_offset()
        self._pause_all_changelog_partitions()
        self._resume_partitions(self._primary_partitions)



from fluvii.general_utils import log_and_raise_error
from fluvii.exceptions import SignalRaise, GracefulTransactionFailure, FatalTransactionFailure, FinishedTransactionBatch, FailedAbort, TransactionCommitted, TransactionNotRequired
from fluvii.consumer import TransactionalConsumer
from fluvii.producer import TransactionalProducer
from fluvii.transaction import Transaction
from fluvii.schema_registry import SchemaRegistry
from fluvii.metrics import MetricsManager
from .config import FluviiConfig
import logging


LOGGER = logging.getLogger(__name__)


class FluviiApp:
    def __init__(self, app_function, consume_topics_list, fluvii_config=None, produce_topic_schema_dict=None, transaction_cls=Transaction,
                 app_function_arglist=None, metrics_manager=None):
        if not app_function_arglist:
            app_function_arglist = []
        if isinstance(consume_topics_list, str):
            consume_topics_list = consume_topics_list.split(',')
        if not produce_topic_schema_dict:  # Should only be the case if you dynamically add topics at runtime...otherwise should just use the plain Consumer
            produce_topic_schema_dict = {}

        self._shutdown = False
        self._config = fluvii_config
        self._transaction_cls = transaction_cls
        self._app_function = app_function
        self._app_function_arglist = app_function_arglist
        self._produce_topic_schema_dict = produce_topic_schema_dict
        self._consume_topics_list = consume_topics_list
        self._consumer = None
        self._producer = None
        self._schema_registry = None

        self.metrics_manager = metrics_manager
        self.transaction = None

        self._set_config()

    def _set_config(self):
        if not self._config:
            self._config = FluviiConfig()

    def _init_clients(self):
        LOGGER.info('Initializing Kafka clients...')
        self._set_schema_registry()
        self._set_producer()
        self._set_consumer()

    def _init_metrics_manager(self):
        LOGGER.info("Initializing the MetricsManager...")
        if not self.metrics_manager:
            self.metrics_manager = MetricsManager(
                metrics_config=self._config.metrics_manager_config, pusher_config=self._config.metrics_pusher_config)

    def _set_schema_registry(self):
        LOGGER.debug('Setting up Schema Registry...')
        self._schema_registry = SchemaRegistry(
            self._config.schema_registry_url,
            auth_config=self._config.schema_registry_auth_config
        ).registry

    def _set_producer(self, force_init=False):
        if (not self._producer) or force_init:
            LOGGER.debug('Setting up Kafka Transactional Producer')
            self._producer = TransactionalProducer(
                urls=self._config.client_urls,
                client_auth_config=self._config.client_auth_config,
                topic_schema_dict=self._produce_topic_schema_dict,
                transactional_id=self._config.hostname,
                metrics_manager=self.metrics_manager,
                schema_registry=self._schema_registry,
                settings_config=self._config.producer_config,
            )
            LOGGER.debug('Producer setup complete.')

    def _set_consumer(self, auto_subscribe=True):
        if not self._consumer:
            self._consumer = TransactionalConsumer(
                urls=self._config.client_urls,
                client_auth_config=self._config.client_auth_config,
                group_id=self._config.app_name,
                consume_topics_list=self._consume_topics_list,
                auto_subscribe=auto_subscribe,
                metrics_manager=self.metrics_manager,
                schema_registry=self._schema_registry,
                settings_config=self._config.consumer_config,
            )

    def abort_transaction(self):
        try:
            self.transaction.abort_transaction()
        except FailedAbort:
            self._set_producer(force_init=True)

    def _init_transaction_handler(self, **kwargs):
        LOGGER.debug('initing a transaction handler...')
        self.transaction = self._transaction_cls(self._producer, self._consumer, fluvii_app_instance=self, auto_consume=False, **kwargs)
        return self.transaction

    def _handle_message(self, **kwargs):
        self.consume(**kwargs)
        self._app_function(self.transaction, *self._app_function_arglist)

    def _no_message_callback(self):
        self._producer.poll(0)
        LOGGER.debug('No messages!')

    def _finalize_app_batch(self):
        LOGGER.debug('Finalizing transaction batch, if necessary...')
        self.commit()

    def _app_batch_run_loop(self, **kwargs):
        LOGGER.info(f'Consuming {self._config.consumer_config.batch_consume_max_count} messages'
                    f' over {self._config.consumer_config.batch_consume_max_time_seconds} seconds...')
        try:
            while not self._shutdown:
                try:
                    self._handle_message(**kwargs)
                except FinishedTransactionBatch:
                    self._finalize_app_batch()
                    raise TransactionCommitted
        except TransactionCommitted:
            pass
        except TransactionNotRequired:
            self._no_message_callback()
        except GracefulTransactionFailure:
            LOGGER.info("Graceful transaction failure; retrying commit...")
            self._app_batch_run_loop(**kwargs)
        except FatalTransactionFailure:
            LOGGER.info("Fatal transaction failure; aborting the transaction and resetting consumer state...")
            self.abort_transaction()

    def _app_shutdown(self):
        LOGGER.info('App is shutting down...')
        self._shutdown = True
        try:
            self.abort_transaction()
        except:
            pass
        self.kafka_cleanup()

    def _runtime_init(self):
        self._init_metrics_manager()
        self._init_clients()

    def _run(self, **kwargs):
        try:
            while not self._shutdown:
                self._init_transaction_handler()
                self._app_batch_run_loop(**kwargs)
        except SignalRaise:
            LOGGER.info('Shutdown requested via SignalRaise!')
        except Exception as e:
            LOGGER.error(f'{e.__class__}, {e}')
            if self.metrics_manager:
                log_and_raise_error(self.metrics_manager, e)
        finally:
            self._app_shutdown()

    def kafka_cleanup(self):
        """ Public method in the rare cases where you need to do some cleanup on the consumer object manually. """
        LOGGER.info('Performing graceful teardown of producer and/or consumer...')
        if self._consumer:
            LOGGER.debug("Shutting down consumer; no further commits can be queued or finalized.")
            # TODO: make sure consumer unsubscribes on close.
            self._consumer.close()

    def consume(self, **kwargs):
        LOGGER.debug('Calling consume...')
        self.transaction.consume(**kwargs)

    def commit(self):
        LOGGER.debug('Calling commit...')
        self.transaction.commit()

    def run(self, **kwargs):
        LOGGER.info('RUN initialized!')
        self._runtime_init()
        self._run(**kwargs)
        # TODO: consider just returning self._run here to allow easy way to return things?



"""
Container for standardized application monitoring gauges
"""
from prometheus_client import Gauge, CollectorRegistry
from .config import MetricsManagerConfig
from fluvii.metrics.pusher import MetricsPusher, MetricsPusherConfig


class Metric:
    def __init__(self, metric_name, registry, metrics_config=None, host=None, app=None, description='', additional_labels=None):
        if not metrics_config:
            metrics_config = MetricsManagerConfig()

        if not app:
            app = metrics_config.app_name
        if not host:
            host = metrics_config.hostname
        if not additional_labels:
            additional_labels = []

        self.name = metric_name
        self.app = app
        self.host = host
        self.additional_labels = list(sorted(additional_labels))
        self.all_labels = ['app', 'job'] + self.additional_labels
        self.store = Gauge(self.name, description, labelnames=self.all_labels, registry=registry)

    def _labels(self, label_dict):
        if not self.additional_labels:
            label_dict = {}
        return self.store.labels(self.app, self.host, *[i[1] for i in sorted(label_dict.items())])

    def inc(self, number=1, label_dict=None):
        self._labels(label_dict).inc(number)

    def set(self, number, label_dict=None):
        self._labels(label_dict).set(number)


class MetricsManager:
    """
    Coordinates Prometheus monitoring for a Kafka client application

    Creates and manages Metric instances and pushes their metrics
    """

    def __init__(self, metrics_config=MetricsManagerConfig(), registry=CollectorRegistry(), pusher_cls=MetricsPusher, pusher_config=MetricsPusherConfig()):
        """
        Initializes monitor and Metric classes
        """
        self._config = metrics_config
        self.registry = registry
        self._metrics = {}

        self.new_metric('messages_consumed', description='Messages consumed since application start', additional_labels=['topic']),
        self.new_metric('messages_produced', description='Messages produced since application start', additional_labels=['topic']),
        self.new_metric('message_errors', description='Exceptions caught when processing messages', additional_labels=['exception']),
        self.new_metric('external_requests', description='Network calls to external services', additional_labels=['request_to', 'request_endpoint', 'request_type', 'is_bulk', 'status_code']),
        self.new_metric('seconds_behind', description='Elapsed time since the consumed message was originally produced')

        self.pusher = pusher_cls(self.registry, pusher_config)

    def __getattr__(self, name):
        try:
            return self._metrics[name]
        except:
            return super().__getattribute__(name)

    @property
    def metric_names(self):
        return [metric.name for metric in self._metrics.values()]

    def new_metric(self, metric_name, additional_labels=None, **kwargs):
        self._metrics[metric_name] = Metric(metric_name, self.registry, additional_labels=additional_labels, metrics_config=self._config, **kwargs)

    def inc_metric(self, metric_name, number=1, label_dict=None):
        self._metrics[metric_name].inc(number=number, label_dict=label_dict)

    def set_metric(self, metric_name, number, label_dict=None):
        self._metrics[metric_name].set(number, label_dict=label_dict)



from confluent_kafka import KafkaException
import logging
from copy import deepcopy
from fluvii.general_utils import parse_headers
from fluvii.exceptions import GracefulTransactionFailure, FatalTransactionFailure, FailedAbort
from json import dumps, loads


LOGGER = logging.getLogger(__name__)


def handle_kafka_exception(kafka_error):
    LOGGER.error(kafka_error)
    LOGGER.error(kafka_error.args[0].code)
    # if kafka_error.args[0].code == 'ILLEGAL GENERATION':
    #     LOGGER.info('The consumer group generation id is invalid (likely due to a rebalance call), aborting transaction')
    #     raise FatalTransactionFailure
    retriable = kafka_error.args[0].retriable()
    abort = kafka_error.args[0].txn_requires_abort()
    # TODO: take advantage of retriable
    LOGGER.info(f'KafkaException: is retriable? - {retriable}, should abort? - {abort}')
    if retriable:
        raise GracefulTransactionFailure
    elif abort:
        raise FatalTransactionFailure
    else:
        pass


class Transaction:
    def __init__(self, producer, consumer, auto_consume=True, message=None, fluvii_app_instance=None, refresh_after_commit=False):
        self.producer = producer
        self.consumer = consumer
        self.message = message
        self.metrics_manager = self.consumer.metrics_manager
        self._allow_auto_consume = auto_consume
        self._refresh_after_commit = refresh_after_commit  # allows you to re-use the object, as needed
        self._init_attrs()

        # this optional attribute should never be referenced/used by class methods here to keep transactions independent of the app
        # only intended for runtime access of the app instance by the transaction
        self.app = fluvii_app_instance

    def _init_attrs(self):
        self._auto_consume(self.message, self._allow_auto_consume)

    def _auto_consume(self, has_input, should_consume):
        """ Consume message if initialized without one (and allowed to) """
        if not has_input and should_consume:
            self.consume()

    @property
    def has_outstanding_updates(self):
        return self.producer.active_transaction or self.consumer.pending_commits

    def messages(self):  # here for a consistent way to access message(s) currently being managed for when you subclass
        """ For a standardized way to access message(s) consumed pertaining to this transaction """
        return self.consumer.messages()

    def consume(self, **kwargs):
        self.message = self.consumer.consume(**kwargs)

    def key(self):
        return deepcopy(self.message.key())

    def value(self):
        return deepcopy(self.message.value())

    def headers(self):
        return deepcopy(parse_headers(self.message.headers()))

    def topic(self):
        return self.message.topic()

    def partition(self):
        return self.message.partition()

    def offset(self):
        return self.message.offset()

    def _abort_transaction(self):
        try:
            LOGGER.info('Aborting transaction.')
            self.producer.abort_transaction(10)
        except KafkaException as e:
            LOGGER.info(f"Failed to abort transaction: {e}")
            raise FailedAbort

    def abort_transaction(self):
        LOGGER.debug('Aborting any open transactions, if needed.')
        if self.consumer.pending_commits:
            self.consumer.rollback_consumption()
        if self.producer.active_transaction:
            self._abort_transaction()
        self._init_attrs()

    def produce(self, producer_kwargs):
        value = producer_kwargs.pop('value', producer_kwargs)
        self.producer.produce(value, message_passthrough=self.message, **producer_kwargs)

    def _commit(self):
        try:
            self.consumer.commit(self.producer)
        except KafkaException as kafka_error:
            handle_kafka_exception(kafka_error)

    def commit(self):
        """ Allows manual commits (safety measures in place so that you cant commit the same message twice)."""
        self._commit()
        if self._refresh_after_commit:
            self._init_attrs()


class TableTransaction(Transaction):
    def __init__(self, producer, consumer, fluvii_changelog_topic, fluvii_tables, auto_consume=True, message=None, fluvii_app_instance=None, refresh_after_commit=False):
        # set first since we override _init_attrs
        self.app_changelog_topic = fluvii_changelog_topic
        self.tables = fluvii_tables

        super().__init__(producer, consumer, message=message, auto_consume=auto_consume,
                         fluvii_app_instance=fluvii_app_instance, refresh_after_commit=refresh_after_commit)

    # -------------------------  Protected Method Overrides ------------------------
    def _init_attrs(self):
        super()._init_attrs()
        self._is_not_changelog_message = True
        self._pending_table_writes = {p: {} for p in self.tables}
        self._pending_table_offset_increase = {p: 0 for p in self.tables}

    def _commit(self):
        super()._commit()
        self._table_write()

    # ----------------------- Protected Method Extensions -----------------------
    def _table_offset(self, partition=None):
        if partition is None:
            partition = self.partition()
        value = self.tables[partition].offset
        value = int(value) if value else 0
        return value + self._pending_table_offset_increase.get(partition, 0)

    def _update_pending_table_writes(self, value):
        self._pending_table_writes[self.partition()][self.key()] = value
        if self._is_not_changelog_message:
            self._pending_table_offset_increase[self.partition()] += 1
        else:
            self._pending_table_offset_increase[self.partition()] += self.offset() - self._table_offset()

    def _update_changelog(self):
        if self._is_not_changelog_message and (pending_write := self._pending_table_writes.get(self.partition(), {}).get(self.key())):
            LOGGER.debug(f'Updating changelog topic for {self.key()}')
            self.produce(dict(value=dumps(pending_write), topic=self.app_changelog_topic, key=self.key(), partition=self.partition()))

    def _update_table_entry_from_changelog(self):
        self._is_not_changelog_message = False  # so we dont produce a message back to the changelog
        self.update_table_entry(loads(self.value()))

    def _table_write(self, recovery_multiplier=1):
        for p, msgs in self._pending_table_writes.items():
            if msgs:
                table = self.tables[p]
                LOGGER.debug(f'Finalizing table entry batch write of {len(msgs)} records for table {table.table_name}')
                LOGGER.debug(f'Table {table.table_name} offset before write: {table.offset}, expected after write: {self._table_offset(partition=p) + int(self._is_not_changelog_message)}')
                table.set_offset(self._table_offset(partition=p) + int(self._is_not_changelog_message))  # the +1 is because the transaction causes an extra offset at the end (per trans + partition)
                table.write_batch(msgs)
                table.commit_and_cleanup_if_ready(recovery_multiplier=recovery_multiplier)
                self._pending_table_writes[p] = {}
                self._pending_table_offset_increase[p] = 0

    # ----------------------- Method Extensions -----------------------
    def read_table_entry(self):
        pending_update = self._pending_table_writes.get(self.partition(), {}).get(self.key())
        if pending_update:
            return deepcopy(pending_update)
        return self.tables[self.partition()].read(self.key())

    def update_table_entry(self, value):
        self._update_pending_table_writes(value)
        self._update_changelog()

    def delete_table_entry(self):
        self._update_pending_table_writes('-DELETED-')
        self._update_changelog()



from . import FluviiApp


class FluviiMultiMessageApp(FluviiApp):
    """
    This is for when you need to handle multiple messages at once, like for doing a bulk api batch.
    Your app_function should be written with the expectation that transaction.messages() is a list of messages to manipulate
    """
    def _set_config(self):
        super()._set_config()
        self._config.consumer_config.batch_consume_store_messages = True

    def _handle_message(self, **kwargs):
        # Don't do the app function here anymore!
        self.consume(**kwargs)

    def _finalize_app_batch(self):
        # Do it at the end of consuming instead!
        self._app_function(self.transaction, *self._app_function_arglist)
        super()._finalize_app_batch()



from confluent_kafka import DeserializingConsumer, TopicPartition
from confluent_kafka.schema_registry.avro import AvroDeserializer
from fluvii.exceptions import NoMessageError, ConsumeMessageError, FinishedTransactionBatch, TransactionNotRequired
from fluvii.general_utils import parse_headers, get_guid_from_message
from copy import deepcopy
import logging
import datetime

LOGGER = logging.getLogger(__name__)


class Consumer:
    def __init__(self, urls, group_id, consume_topics_list, schema_registry=None, auto_subscribe=True, client_auth_config=None, settings_config=None, metrics_manager=None, consumer_cls=DeserializingConsumer):
        self._urls = ','.join(urls) if isinstance(urls, list) else urls
        self._auth = client_auth_config
        self._settings = settings_config
        self._consumer = None
        self._group_id = group_id
        self._topic_metadata = None
        self._schema_registry = schema_registry
        self._consumer_cls = consumer_cls

        self.message = None
        self.metrics_manager = metrics_manager
        self.topics = consume_topics_list if isinstance(consume_topics_list, list) else consume_topics_list.split(',')
        self._poll_timeout = self._settings.poll_timeout_seconds if self._settings else 5

        self._init_consumer(auto_subscribe=auto_subscribe)

    def __getattr__(self, attr):
        """Note: this includes methods as well!"""
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            return self._consumer.__getattribute__(attr)

    @staticmethod
    def _consume_message_callback(error, partitions):
        """
        Logs the info returned when a successful commit is performed
        NOTE: Callback requires these args
        """
        if error:
            LOGGER.critical(error)
        else:
            LOGGER.debug('Consumer Callback - Message consumption committed successfully')

    def _make_config(self):
        settings = {
            "bootstrap.servers": self._urls,

            "group.id": self._group_id,
            "on_commit": self._consume_message_callback,
            "enable.auto.offset.store": False,  # ensures auto-committing doesn't happen before the consumed message is actually finished
            "partition.assignment.strategy": 'cooperative-sticky',

            # Registry Serialization Settings
            "key.deserializer": AvroDeserializer(self._schema_registry, schema_str='{"type": "string"}'),
            "value.deserializer": AvroDeserializer(self._schema_registry) if self._schema_registry else None,
        }

        if self._settings:
            settings.update(self._settings.as_client_dict())
        if self._auth:
            settings.update(self._auth.as_client_dict())
        return settings

    def _init_consumer(self, auto_subscribe=True):
        LOGGER.info('Initializing Consumer...')
        self._consumer = self._consumer_cls(self._make_config())
        LOGGER.info('Consumer Initialized!')
        if auto_subscribe:
            self._consumer.subscribe(topics=self.topics)
            LOGGER.info(f'Consumer subscribed to topics {self.topics}!')

    def _poll_for_message(self, timeout=None):
        """Is a separate method to isolate communication interface with Kafka"""
        if not timeout:
            timeout = self._poll_timeout
        message = self._consumer.poll(timeout)
        if message is None:
            raise NoMessageError
        return message

    def _handle_consumed_message(self):
        """
        Handles a consumed message to check for errors and log the consumption as a metric.
        If the message is returned with a breaking error, raises a ConsumeMessageError.
        """
        try:
            guid = get_guid_from_message(self.message)
            if '__changelog' not in self.message.topic():
                LOGGER.info(f"Message consumed from topic {self.message.topic()} partition {self.message.partition()}, offset {self.message.offset()}; GUID {guid}")
                LOGGER.debug(f"Consumed message key: {repr(self.message.key())}")
                if self.metrics_manager:
                    self.metrics_manager.set_metric('seconds_behind', round(datetime.datetime.timestamp(datetime.datetime.utcnow())) - self.message.timestamp()[1] // 1000)
                    self.metrics_manager.inc_metric('messages_consumed', label_dict={'topic': self.message.topic()})
        except AttributeError:
            if "object has no attribute 'headers'" in str(self.message.error()):
                raise ConsumeMessageError("Headers were inaccessible on the message. Potentially a corrupt message?")

    def key(self):
        return deepcopy(self.message.key())

    def value(self):
        return deepcopy(self.message.value())

    def headers(self):
        return deepcopy(parse_headers(self.message.headers()))

    def messages(self):
        return [self.message]

    def consume(self, timeout=None):
        """
        Consumes a message from the broker while handling errors.
        If the message is valid, then the message is returned.
        """
        self.message = self._poll_for_message(timeout)
        self._handle_consumed_message()
        return self.message

    def commit(self):
        self._consumer.store_offsets(self.message)
        self.message = None


class TransactionalConsumer(Consumer):
    def __init__(self, urls, group_id, consume_topics_list, schema_registry=None, auto_subscribe=True,
                 client_auth_config=None, settings_config=None, metrics_manager=None,
                 ):
        super().__init__(urls, group_id, consume_topics_list, schema_registry=schema_registry, auto_subscribe=auto_subscribe,
                         client_auth_config=client_auth_config, settings_config=settings_config, metrics_manager=metrics_manager)
        self._batch_time_elapse_start = None
        self._consume_max_time_secs = self._settings.batch_consume_max_time_seconds
        self._consume_max_count = self._settings.batch_consume_max_count
        self._consume_max_empty_polls = self._settings.batch_consume_max_empty_polls
        self._store_batch_messages = self._settings.batch_consume_store_messages
        self._init_attrs()
        self._reset_keep_consuming_trackers()

    def _reset_keep_consuming_trackers(self):
        self._batch_time_elapse_start = None
        self._batch_remaining_empty_polls = self._consume_max_empty_polls
        self._consume_message_count = 0

    def _init_attrs(self):
        self._batch_offset_starts = {}
        self._batch_offset_ends = {}
        self._messages = []
        self.message = None
    
    def _set_batch_start_time(self):
        self._batch_time_elapse_start = datetime.datetime.now().timestamp()

    def _max_consume_time_continue(self):
        continue_consume = True
        if self._consume_max_time_secs:
            if not self._batch_time_elapse_start:
                self._set_batch_start_time()
            seconds_elapsed = datetime.datetime.now().timestamp() - self._batch_time_elapse_start
            continue_consume = seconds_elapsed < self._consume_max_time_secs
            if not continue_consume:
                self._batch_time_elapse_start = None
        return continue_consume

    def _max_consume_count_continue(self, consume_multiplier=1):
        if self._consume_max_count:
            return self._consume_message_count < (self._consume_max_count * consume_multiplier)
        return True

    def _keep_consuming(self, consume_multiplier=1):
        return self._batch_remaining_empty_polls and self._max_consume_count_continue(consume_multiplier=consume_multiplier) and self._max_consume_time_continue()

    def _mark_offset_start(self):
        if self.message.topic() not in self._batch_offset_starts:
            self._batch_offset_starts[self.message.topic()] = {}
        if self.message.partition() not in self._batch_offset_starts[self.message.topic()]:
            self._batch_offset_starts[self.message.topic()][self.message.partition()] = self.message.offset()

    def _mark_offset_end(self):
        if self.message.topic() not in self._batch_offset_ends:
            self._batch_offset_ends[self.message.topic()] = {}
        self._batch_offset_ends[self.message.topic()][self.message.partition()] = self.message.offset()

    def _get_consumer_partition_assignment(self):
        assignments = self._consumer.assignment()
        assignments = {topic: [int(obj.partition) for obj in assignments if obj.topic == topic] for topic in set([obj.topic for obj in assignments])}
        return assignments

    def _commit(self, producer, offsets):
        if offsets:
            LOGGER.debug(f'Offset ends ready for commit: {self._batch_offset_ends}')
            if not producer.active_transaction:
                producer.begin_transaction()
            producer.send_offsets_to_transaction(offsets, self._consumer.consumer_group_metadata())
        else:
            LOGGER.info('No messages were consumed in this batch.')

        if producer.active_transaction:
            LOGGER.info('Comitting transaction!')
            producer.commit_transaction(30)
        else:
            raise TransactionNotRequired

    def _make_config(self):
        config = super()._make_config()
        config.update({
            "isolation.level": "read_committed",
            "enable.auto.commit": False,
            "enable.auto.offset.store": False,
        })
        return config

    def _handle_consumed_message(self):
        super()._handle_consumed_message()
        self._mark_offset_start()
        self._mark_offset_end()
        self._consume_message_count += 1
        if self._store_batch_messages:
            self._messages.append(self.message)

    @property
    def pending_commits(self):
        return bool(self._mark_offset_end)

    def messages(self):
        if self._store_batch_messages:
            return self._messages
        return super().messages()

    def rollback_consumption(self):
        LOGGER.info('Rolling back consumer state to earliest non-committed offset(s)...')
        assignments = self._get_consumer_partition_assignment()
        for topic, partitions in self._batch_offset_starts.items():
            for partition, offset in partitions.items():
                if partition in assignments.get(topic, []):
                    LOGGER.info(f"Reversing topic {topic} partition {partition} back to offset {offset}")
                    self._consumer.seek(TopicPartition(topic=topic, partition=partition, offset=offset))
        self._init_attrs()
        self._reset_keep_consuming_trackers()

    def commit(self, producer):
        offsets_to_commit = [TopicPartition(topic, partition, offset + 1) for topic, partitions in
                             self._batch_offset_ends.items() for partition, offset in partitions.items()]
        self._commit(producer, offsets_to_commit)
        self._init_attrs()

    def consume(self, timeout=None, consume_multiplier=1):
        """
        Consumes a message from the broker while handling errors.
        If the message is valid, then the message is returned.
        """
        while self._keep_consuming(consume_multiplier=consume_multiplier):
            try:
                return super().consume(timeout=timeout)
            except NoMessageError:
                self._batch_remaining_empty_polls -= 1
        LOGGER.info('Consumption attempts for this batch are finished.')
        self._reset_keep_consuming_trackers()
        raise FinishedTransactionBatch



from os import environ
from typing import Literal, Optional

from pydantic import BaseSettings, Field

from fluvii.config_base import KafkaConfigBase


class ConsumerConfig(KafkaConfigBase, BaseSettings):
    """
    Common configs, along with some custom ones, that likely wont need to be changed from their defaults.
    """
    auto_commit_interval_seconds: int = 20
    auto_offset_reset: Literal["earliest", "latest"] = "latest"
    batch_consume_max_count: int = 100
    batch_consume_max_empty_polls: Optional[int] = 2
    batch_consume_max_time_seconds: Optional[int] = 10
    batch_consume_store_messages: bool = False
    heartbeat_timeout_ms: int = 4 * 60 // 2 * 1_000  # TODO document that this pairs with timeout_minutes
    message_singleton_max_mb: int = 2
    message_batch_max_mb: int = 5
    message_queue_max_mb: int = 20
    poll_timeout_seconds: int = 5
    timeout_minutes: int = 4  # TODO document that this pairs with heartbeat_timeout_ms

    class Config:
        env_prefix = "FLUVII_CONSUMER_"

    def as_client_dict(self):
        ms_tolerance = 1_000
        return {
            "auto.commit.interval.ms": self.auto_commit_interval_seconds * 1_000,
            "auto.offset.reset": self.auto_offset_reset,
            "fetch.max.bytes": self.message_batch_max_mb * (2 ** 20),
            "heartbeat.interval.ms": (self.heartbeat_timeout_ms // 5) - ms_tolerance,  # 5 failed heartbeats == bad consumer.
            "max.poll.interval.ms": self.timeout_minutes * 60_000,  # Max time between poll() calls before considered dead.
            "message.max.bytes": self.message_singleton_max_mb * (2 ** 20),
            "queued.max.messages.kbytes": self.message_queue_max_mb * (2 ** 10),
            "session.timeout.ms": self.heartbeat_timeout_ms,  # need at least 1 heartbeat within "session" time to be considered alive;
        }



from fluvii.config_base import KafkaConfigBase
import requests
import time


class SaslPlainClientConfig(KafkaConfigBase):
    def __init__(self, username, password):
        self.username = username
        self.password = password

    def as_client_dict(self):
        return {
            "security.protocol": 'SASL_SSL',
            "sasl.mechanisms": 'PLAIN',
            "sasl.username": self.username,
            "sasl.password": self.password,
        }


class SaslOauthClientConfig(KafkaConfigBase):
    def __init__(self, username, password, url, scope):
        self.username = username
        self.password = password
        self.url = url
        self.scope = scope

    def _get_token(self, required_arg):
        """required_arg is...well, required. Was easier to set it up without using it (basically
        is just passed whatever you set sasl.oauthbearer.config to...(on the client, I'm assuming?))"""
        payload = {
            'grant_type': 'client_credentials',
            'scope': self.scope
        }
        resp = requests.post(self.url,
                             auth=(self.username, self.password),
                             data=payload)
        token = resp.json()
        return token['access_token'], time.time() + float(token['expires_in'])

    def as_client_dict(self):
        return {
            'security.protocol': 'SASL_SSL',
            'sasl.mechanisms': 'OAUTHBEARER',
            'oauth_cb': self._get_token,
        }



from time import sleep
from confluent_kafka.admin import AdminClient


class Admin:
    def __init__(self, urls, auth=None):
        if not auth:
            auth = {}
        self._auth = auth
        self._urls = urls
        self._admin_client = None
        self._init_admin_client()

    def __getattr__(self, attr):
        """Note: this includes methods as well!"""
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            return self._admin_client.__getattribute__(attr)

    def _init_admin_client(self):
        auth = self._auth.as_client_dict() if self._auth else self._auth
        self._admin_client = AdminClient({
            "bootstrap.servers": self._urls,
            **auth})


def parse_headers(msg_header):
    """
    Converts headers to a dict
    :param msg_header: A message .headers() (confluent)
    :return: a decoded dict version of the headers
    """
    if msg_header:
        if not isinstance(msg_header, dict):
            msg_header = dict(msg_header)
            return {key: value.decode() for key, value in msg_header.items()}
        return msg_header
    return {}


def get_guid_from_message(message):
    guid = None
    headers = (i for i in message.headers())
    while not guid:
        h = next(headers)
        guid = h[1].decode() if h[0] == 'guid' else None
    return guid


def log_and_raise_error(metrics_manager, error):
    """
    Since the metric manager pushing is a separate thread, ensure an exception gets sent to prometheus
    """
    metrics_manager.inc_message_errors(error)
    sleep(5)
    raise




from os import environ


class MetricsPusherConfig:
    def __init__(self, hostname=None):
        if not hostname:
            hostname = environ.get('FLUVII_HOSTNAME')
        self.hostname = hostname
        self.headless_service_name = environ.get('FLUVII_METRICS_PUSHER_KUBERNETES_HEADLESS_SERVICE_NAME')
        self.headless_service_port = environ.get('FLUVII_METRICS_PUSHER_KUBERNETES_HEADLESS_SERVICE_PORT')
        self.metrics_port = environ.get('FLUVII_METRICS_PUSHER_KUBERNETES_POD_APP_PORT')
        self.push_rate_seconds = int(environ.get('FLUVII_METRICS_PUSHER_PUSH_RATE_SECONDS', '10'))

        self.enable_pushing = environ.get('FLUVII_METRICS_PUSHER_ENABLE_PUSHING', 'false')
        if self.enable_pushing.lower() == 'true':
            self.enable_pushing = True
        else:
            self.enable_pushing = False



from fluvii.exceptions import PartitionsAssigned, FinishedTransactionBatch, GracefulTransactionFailure, FatalTransactionFailure, TransactionCommitted, TransactionNotRequired
from fluvii.transaction import TableTransaction
from .fluvii_app import FluviiApp
from .rebalance_manager import TableRebalanceManager
from datetime import datetime
import logging


LOGGER = logging.getLogger(__name__)


class FluviiTableApp(FluviiApp):
    def __init__(self, app_function, consume_topic, fluvii_config=None, produce_topic_schema_dict=None, transaction_cls=TableTransaction,
                 app_function_arglist=None, metrics_manager=None, table_changelog_topic=None, table_recovery_multiplier=None):

        super().__init__(
            app_function, consume_topic, fluvii_config=fluvii_config, produce_topic_schema_dict=produce_topic_schema_dict,
            transaction_cls=transaction_cls, app_function_arglist=app_function_arglist, metrics_manager=metrics_manager)

        if not table_changelog_topic:
            table_changelog_topic = self._config.table_changelog_topic
        self.topic = consume_topic
        self.changelog_topic = table_changelog_topic
        self.tables = {}
        self._rebalance_manager = None

        # # Table recovery attributes
        if not table_recovery_multiplier:
            table_recovery_multiplier = self._config.table_recovery_multiplier
        self._recovery_multiplier = table_recovery_multiplier

    # -------------------------  Protected Method Overrides ------------------------
    def _set_consumer(self):
        super()._set_consumer(auto_subscribe=False)
        self._consumer.subscribe(self._consume_topics_list, on_assign=self._partition_assignment,
                                 on_revoke=self._partition_unassignment, on_lost=self._partition_unassignment)

    def _init_transaction_handler(self, **kwargs):
        super()._init_transaction_handler(fluvii_changelog_topic=self.changelog_topic, fluvii_tables=self.tables, **kwargs)

    def _finalize_app_batch(self):
        super()._finalize_app_batch()
        self.check_table_commits()

    def _no_message_callback(self):
        super()._no_message_callback()
        self.commit_tables()

    def _app_batch_run_loop(self, **kwargs):
        try:
            super()._app_batch_run_loop(**kwargs)
        except PartitionsAssigned:
            self._handle_rebalance()

    def _runtime_init(self):
        super()._runtime_init()
        self._producer.add_topic(self.changelog_topic, {"type": "string"})
        self._rebalance_manager = TableRebalanceManager(self._consumer, self.changelog_topic, self.tables, self._config)

    def _app_shutdown(self):
        super()._app_shutdown()

    # ----------------------- Protected Method Extensions -----------------------
    def _init_recovery_time_remaining_attrs(self):
        self._recovery_time_start = int(datetime.timestamp(datetime.now()))
        self._recovery_offsets_remaining = 0
        self._recovery_offsets_handled = 0

    def _partition_assignment(self, consumer, add_partition_objs):
        """
        Called every time a rebalance happens and handles table assignment and recovery flow.
        NOTE: confluent-kafka expects this method to have exactly these two arguments ONLY
        NOTE: _partition_assignment will ALWAYS be called (even when no new assignments are required) after _partition_unassignment.
            """
        if not self._shutdown:
            LOGGER.debug('Rebalance Triggered - Assigment')
            self.abort_transaction()
            self._rebalance_manager.assign_partitions(add_partition_objs)
            raise PartitionsAssigned

    def _partition_unassignment(self, consumer, drop_partition_objs):
        """
        NOTE: confluent-kafka expects this method to have exactly these two arguments ONLY
        NOTE: _partition_assignment will always be called (even when no new assignments are required) after _partition_unassignment.
        """
        LOGGER.info('Rebalance Triggered - Unassigment')
        if not self._shutdown:
            self.abort_transaction()
        self._rebalance_manager.unassign_partitions(drop_partition_objs)

    def _handle_recovery_message(self):
        self.transaction.consume(consume_multiplier=self._recovery_multiplier)
        self.transaction._update_table_entry_from_changelog()

    def _finalize_recovery_batch(self):
        if self.transaction._pending_table_writes:
            self.transaction._table_write(recovery_multiplier=self._recovery_multiplier)  # relay transaction's cached writes to the table's write cache
            self._consumer._init_attrs()
        else:
            raise TransactionNotRequired

    def _table_recovery_consume_loop(self, checks):
        LOGGER.info(f'Consuming from changelog partitions: {[p.partition for p in self._rebalance_manager.recovery_partitions]}')
        LOGGER.info(f'Processing up to {self._config.consumer_config.batch_consume_max_count * self._recovery_multiplier} messages for up to {self._config.consumer_config.batch_consume_max_time_seconds} seconds!')
        # NOTE: no transaction commits since its just consuming from changelog and writing to the table, we dont care about the consumer group offset
        try:
            while not self._shutdown:
                try:
                    self._handle_recovery_message()
                except FinishedTransactionBatch:
                    self._finalize_recovery_batch()
                    raise TransactionCommitted
        except TransactionCommitted:
            self._rebalance_manager.update_recovery_status()
        except TransactionNotRequired:
            checks -= 1
            LOGGER.info(f'No further changelog messages, checks remaining: {checks}')
            self.commit_tables()
            self._rebalance_manager.update_recovery_status()
        except GracefulTransactionFailure:
            self._finalize_app_batch()
            self._table_recovery_consume_loop(checks)
        except FatalTransactionFailure:
            self.abort_transaction()
            self._table_recovery_consume_loop(checks)
        return checks

    def _handle_rebalance(self):
        try:
            LOGGER.debug('Rebalance management initiated...')
            self._rebalance_manager.set_table_and_recovery_state()
            if self._rebalance_manager.recovery_partitions:
                LOGGER.info('BEGINNING TABLE RECOVERY PROCEDURE')
                checks = 2
                while checks and self._rebalance_manager.recovery_partitions:
                    self._init_transaction_handler()
                    checks = self._table_recovery_consume_loop(checks)
                self.commit_tables()
                LOGGER.info('TABLE RECOVERY COMPLETE!')
            else:
                LOGGER.info('NO TABLE RECOVERY REQUIRED!')
            self._rebalance_manager.finalize_rebalance()
        except PartitionsAssigned:
            self._handle_rebalance()

    # ----------------------- Method Extensions -----------------------
    def check_table_commits(self, recovery_multiplier=1):
        for table in self.tables.values():
            table.commit_and_cleanup_if_ready(recovery_multiplier=recovery_multiplier)

    def commit_tables(self):
        for table in self.tables.values():
            table.commit()



from fluvii.config_base import KafkaConfigBase
from os import environ


class ProducerConfig(KafkaConfigBase):
    """
    Common configs, along with some custom ones, that likely wont need to be changed from their defaults.
    """
    def __init__(self):
        self.transaction_timeout_mins = int(environ.get('FLUVII_TRANSACTION_TIMEOUT_MINUTES', '1'))

    def as_client_dict(self):
        return {
            "transaction.timeout.ms": self.transaction_timeout_mins * 60000,
        }



from fluvii.producer import ProducerConfig
from fluvii.consumer import ConsumerConfig
from fluvii.auth import SaslPlainClientConfig, SaslOauthClientConfig
from fluvii.metrics import MetricsManagerConfig, MetricsPusherConfig
from os import environ
from datetime import datetime
import logging

LOGGER = logging.getLogger(__name__)


class FluviiConfig:
    """
    Manages configuration setup for FLUVII applications

    Generally, is getting things from environment variables.
    Will default to values for certain optional variables,
    and raise on exception for variables that are required.
    """
    def __init__(self,
                 client_urls=None, schema_registry_url=None, client_auth_config=None, schema_registry_auth_config=None,
                 producer_config=None, consumer_config=None, metrics_manager_config=None, metrics_pusher_config=None):
        self._time = int(datetime.timestamp(datetime.now()))

        # Set via properties
        self._hostname = None
        self._table_changelog_topic = None

        # Required vars
        if not client_urls:
            client_urls = environ['FLUVII_KAFKA_BOOTSTRAP_SERVERS']
        if not schema_registry_url:
            schema_registry_url = environ['FLUVII_SCHEMA_REGISTRY_URL']
        self.client_urls = client_urls
        self.schema_registry_url = schema_registry_url

        # Set only if env var or object is specified
        if not client_auth_config:
            if environ.get("FLUVII_CLIENT_USERNAME"):
                if environ.get("FLUVII_OAUTH_URL"):
                    LOGGER.info('Kafka clients will be initialized with OAUTH authentication')
                    client_auth_config = SaslOauthClientConfig(
                        environ["FLUVII_CLIENT_USERNAME"],
                        environ["FLUVII_CLIENT_PASSWORD"],
                        environ["FLUVII_OAUTH_URL"],
                        environ["FLUVII_OAUTH_SCOPE"])
                else:
                    LOGGER.info('Kafka clients will be initialized with plain authentication')
                    client_auth_config = SaslPlainClientConfig(
                        environ["FLUVII_CLIENT_USERNAME"],
                        environ["FLUVII_CLIENT_PASSWORD"])
        if not schema_registry_auth_config:
            if environ.get("FLUVII_SCHEMA_REGISTRY_USERNAME"):
                schema_registry_auth_config = SaslPlainClientConfig(
                    environ["FLUVII_SCHEMA_REGISTRY_USERNAME"],
                    environ["FLUVII_SCHEMA_REGISTRY_PASSWORD"])
        self.client_auth_config = client_auth_config
        self.schema_registry_auth_config = schema_registry_auth_config

        # Everything else with defaults
        if not producer_config:
            producer_config = ProducerConfig()
        if not consumer_config:
            consumer_config = ConsumerConfig()

        self.app_name = environ.get("FLUVII_APP_NAME", 'fluvii_app')
        self.table_folder_path = environ.get('FLUVII_TABLE_FOLDER_PATH', '/tmp')
        self.table_recovery_multiplier = int(environ.get('FLUVII_TABLE_RECOVERY_MULTIPLIER', '10'))
        self.loglevel = environ.get('FLUVII_LOGLEVEL', 'INFO')

        self._hostname = None
        self._table_changelog_topic = None

        # TODO: come up with a way to handle config settings that overlap multiple config objects, like w/metrics configs
        if not metrics_manager_config:
            metrics_manager_config = MetricsManagerConfig(hostname=self.hostname, app_name=self.app_name)
        if not metrics_pusher_config:
            metrics_pusher_config = MetricsPusherConfig(hostname=self.hostname)
        self.consumer_config = consumer_config
        self.producer_config = producer_config
        self.metrics_manager_config = metrics_manager_config
        self.metrics_pusher_config = metrics_pusher_config

    # These properties allow settings that depend on others to dynamically adjust their defaults
    @property
    def hostname(self):
        if self._hostname:
            return self._hostname
        return environ.get('FLUVII_HOSTNAME', f'{self.app_name}_{self._time}')

    @hostname.setter
    def hostname(self, value):
        self._hostname = value

    @property
    def table_changelog_topic(self):
        if self._table_changelog_topic:
            return self._table_changelog_topic
        return environ.get('FLUVII_TABLE_CHANGELOG_TOPIC', f'{self.app_name}__changelog')

    @table_changelog_topic.setter
    def table_changelog_topic(self, value):
        self._table_changelog_topic = value



import sys
import logging
from os import environ


def _init_handler(loglevel=None):
    if not loglevel:
        loglevel = _get_loglevel()
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(loglevel)

    handler.setFormatter(
        logging.Formatter(
            "PID {process} - {asctime} - {name} - {levelname}: {message}",
            style="{"
        )
    )
    return handler


def _get_loglevel():
    return environ.get('FLUVII_LOGLEVEL', 'INFO')


def init_logger(name, loglevel=None):
    if not loglevel:
        loglevel = _get_loglevel()
    logger = logging.getLogger(name)
    logger.setLevel(loglevel)
    logger.addHandler(_init_handler(loglevel=loglevel))
    logger.propagate = False
    return logger



from os import environ


# TODO: Not happy with how this is organized/works, but needed to release something for now.
class MetricsManagerConfig:
    def __init__(self, hostname=None, app_name=None):

        if not hostname:
            hostname = environ.get('FLUVII_HOSTNAME')
        if not app_name:
            app_name = environ.get('FLUVII_APP_NAME')

        self.hostname = hostname
        self.app_name = app_name



import setuptools


install_requires = [
    "confluent-kafka[avro]>=1.9,<2",
    "mmh3",
    "prometheus_client",
    "pydantic",
    "sqlitedict",
]

dev_requires = install_requires + [
    "pip-tools",
    "pytest",
    "pytest-cov",
    "time-machine",
    "twine",
    "sphinx",
    "sphinx-autodoc-typehints",
]

packages = setuptools.find_packages()

setuptools.setup(
    name="fluvii",
    version="0.2.9",
    description="A simple Kafka streams implementation in Python using confluent-kafka-python",
    packages=packages,
    install_requires=install_requires,
    dev_requires=dev_requires,
    include_package_data=True,
    extras_require={"dev": dev_requires},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



"""
Class for pushing metrics to Prometheus metrics cache (pushgateway)
"""
import logging
import socket
from prometheus_client import push_to_gateway
import threading
import time
from .config import MetricsPusherConfig


LOGGER = logging.getLogger(__name__)


class MetricsPusher:
    """
    Pushes metrics to a prometheus pushgateway in a Kubernetes environment
    """

    def __init__(self, registry, config=MetricsPusherConfig()):
        self._config = config
        self.registry = registry
        self.push_thread = None

        self.metrics_pod_ips = []

        if self._config.enable_pushing:
            self.start_pushing_metrics()

    def _set_metrics_pod_ips(self):
        """
        Queries metrics service for gateway IP addresses

        A single Kubernetes service redirects to multiple IP addresses for
        redundant Prometheus pushgateways.
        :return: None
        """
        try:
            socket_info_list = socket.getaddrinfo(self._config.headless_service_name, self._config.headless_service_port)
            self.metrics_pod_ips = {f'{result[-1][0]}:{self._config.metrics_port}' for result in socket_info_list}
            LOGGER.debug(f'Set gateway addresses: {self.metrics_pod_ips}')
        except Exception:
            LOGGER.exception('Failed to set metric pod ips')

    def _push_metrics(self):
        for gateway in self.metrics_pod_ips:
            try:
                push_to_gateway(gateway, job=self._config.hostname, registry=self.registry, timeout=15)
            except Exception as error:
                LOGGER.error(f'Failed to push to pushgateway {gateway}\n{error}')
                self._set_metrics_pod_ips()

    def _push_metrics_loop(self):
        while True:
            self._set_metrics_pod_ips()
            time.sleep(self._config.push_rate_seconds)
            self._push_metrics()

    def _create_metrics_pushing_thread(self):
        self.push_thread = threading.Thread(target=self._push_metrics_loop, daemon=True)

    def start_pushing_metrics(self):
        if not self.push_thread:
            self._create_metrics_pushing_thread()
        try:
            self.push_thread.start()
        except:
            pass

    def stop_pushing_metrics(self):
        if self.push_thread:
            try:
                self.push_thread.stop()
            except:
                pass



from abc import ABC, abstractmethod


class KafkaConfigBase(ABC):
    """
    Manages the connection configuration for a Kafka client

    The options that need to be defined for a kafka client differ depending
    on which security protocol is being used
    """

    @abstractmethod
    def as_client_dict(self):
        pass



from confluent_kafka.schema_registry import SchemaRegistryClient
from urllib.parse import urlparse


# fixes a bug in confluent-kafka TODO: keep a look out for this in >1.8.2, should be fixed since I took it from a MR.
def patched_schema_loads(schema_str):
    from confluent_kafka.schema_registry.avro import Schema
    schema_str = schema_str.strip()
    if schema_str[0] != "{" and schema_str[0] != "[":
        schema_str = '{"type":' + schema_str + '}'
    return Schema(schema_str, schema_type='AVRO')


import confluent_kafka
confluent_kafka.schema_registry.avro._schema_loads = patched_schema_loads
import logging

LOGGER = logging.getLogger(__name__)


class SchemaRegistry:
    def __init__(self, url, auth_config=None):
        self.registry = None
        self.url = url
        self._auth = auth_config
        self._init_registry()

    def __getattr__(self, attr):
        """Note: this includes methods as well!"""
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            return self.registry.__getattribute__(attr)

    def _init_registry(self):
        url = urlparse(self.url)
        if self._auth:
            auth = f"{self._auth.username}:{self._auth.password}@"
        else:
            auth = ''
        scheme = url.scheme
        if not scheme:
            if auth:
                scheme = 'https://'
            else:
                scheme = 'http://'
            url = url._replace(scheme='')
        self.registry = SchemaRegistryClient({'url': f'{scheme}{auth}{url.geturl()}'})
        LOGGER.info('Registry client initialized successfully!')
